@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 2020-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 2020-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -24,6 +24,7 @@
 #include "mem_mgr/video_mem.h"
 #include "gpu/mem_mgr/mem_desc.h"
 #include "gpu/mem_mgr/heap.h"
+#include "gpu/mem_mgr/phys_mem_allocator/phys_mem_allocator.h"
 #include "gpu/mem_mgr/mem_mgr.h"
 #include "gpu/mem_mgr/mem_utils.h"
 #include "gpu/mem_sys/kern_mem_sys.h"
@@ -125,9 +126,9 @@ _vidmemPmaAllocate
     NV_MEMORY_ALLOCATION_PARAMS *pAllocData     = pAllocRequest->pUserParams;
     OBJGPU                      *pGpu           = pAllocRequest->pGpu;
     MemoryManager               *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
-    PMA                         *pPma           = &pHeap->pmaObject;
+    PMA                         *pPma           = pHeap->pPmaObject;
     NvU64                        size           = 0;
-    NvU32                        pageCount;
+    NvU32                        pageCount      = 0;
     NvU32                        pmaInfoSize;
     NvU64                        pageSize;
     NV_STATUS                    status;
@@ -265,7 +266,19 @@ _vidmemPmaAllocate
     allocOptions.flags |= PMA_ALLOCATE_FORCE_ALIGNMENT;
     allocOptions.alignment = NV_MAX(sizeAlign, pageSize);
 
+    if (FLD_TEST_DRF(OS32, _ATTR2, _ENABLE_LOCALIZED_MEMORY, _UGPU0, pAllocData->attr2))
+    {
+        allocOptions.flags |= PMA_ALLOCATE_LOCALIZED_UGPU0;
+    }
+    else if (FLD_TEST_DRF(OS32, _ATTR2, _ENABLE_LOCALIZED_MEMORY, _UGPU1, pAllocData->attr2))
+    {
+        allocOptions.flags |= PMA_ALLOCATE_LOCALIZED_UGPU1;
+    }
+
     // Get the number of pages to be allocated by PMA
+    NV_CHECK_OR_RETURN(LEVEL_ERROR,
+        (NV_DIV_AND_CEIL(size, pageSize) <= NV_U32_MAX),
+        NV_ERR_NO_MEMORY);
     pageCount = (NvU32) NV_DIV_AND_CEIL(size, pageSize);
 
 retry_alloc:
@@ -277,7 +290,12 @@ retry_alloc:
     }
     else
     {
-        pmaInfoSize = sizeof(PMA_ALLOC_INFO) + ((pageCount - 1) * sizeof(NvU64));
+        NV_CHECK_OR_RETURN(LEVEL_ERROR,
+            portSafeMulU32((pageCount - 1), (sizeof(NvU64)), &pmaInfoSize),
+            NV_ERR_NO_MEMORY);
+        NV_CHECK_OR_RETURN(LEVEL_ERROR,
+            portSafeAddU32(pmaInfoSize, (sizeof(PMA_ALLOC_INFO)), &pmaInfoSize),
+            NV_ERR_NO_MEMORY);
     }
 
     // Alloc the tracking structure and store the values in it.
@@ -355,7 +373,7 @@ vidmemPmaFree
     NvU32           flags
 )
 {
-    PMA   *pPma  = &pHeap->pmaObject;
+    PMA   *pPma  = pHeap->pPmaObject;
     NvU32 pmaFreeFlags = flags;
 
     NV_ASSERT_OR_RETURN_VOID(NULL != pPmaAllocInfo);
@@ -443,20 +461,45 @@ vidmemCopyConstruct
     RS_RES_ALLOC_PARAMS_INTERNAL    *pParams
 )
 {
-    Memory    *pMemorySrc            = dynamicCast(pParams->pSrcRef->pResource, Memory);
-    OBJGPU    *pGpu                  = pMemorySrc->pGpu;
-    NV_STATUS  status;
+    Memory            *pMemorySrc   = dynamicCast(pParams->pSrcRef->pResource, Memory);
+    OBJGPU            *pGpu         = pMemorySrc->pGpu;
+    MEMORY_DESCRIPTOR *pMemDesc     = pMemorySrc->pMemDesc;
+    NV_STATUS          status       = NV_ERR_INVALID_ARGUMENT;
+    KernelBus         *pKernelBus    = GPU_GET_KERNEL_BUS(pGpu);
 
-    NV_ASSERT_OR_RETURN(!memdescGetCustomHeap(pMemorySrc->pMemDesc), NV_ERR_INVALID_ARGUMENT);
+    // No flags specified on the initial static BAR1 mapping 
+    status = kbusIncreaseStaticBar1Refcount_HAL(pGpu, pKernelBus,
+                   pMemDesc, BUS_MAP_FB_FLAGS_NONE);
 
-    SLI_LOOP_START(SLI_LOOP_FLAGS_BC_ONLY)
-        MEMORY_DESCRIPTOR *pSrcMemDesc = memdescGetMemDescFromGpu(pMemorySrc->pMemDesc, pGpu);
-        status = heapReference(pGpu, pSrcMemDesc->pHeap, pMemorySrc->HeapOwner,
-                               pSrcMemDesc);
-        NV_ASSERT(status == NV_OK);
-    SLI_LOOP_END
+    if (status == NV_OK)
+    {
+        // nothing
+    }
+    else if (status == NV_ERR_NOT_SUPPORTED)
+    {
+        status = NV_OK;
+    }
+    else
+    {
+        return status;
+    }
 
-    return NV_OK;
+    switch (memdescGetCustomHeap(pMemDesc))
+    {
+        case MEMDESC_CUSTOM_HEAP_NONE:
+            SLI_LOOP_START(SLI_LOOP_FLAGS_BC_ONLY)
+                MEMORY_DESCRIPTOR *pSrcMemDesc = memdescGetMemDescFromGpu(pMemorySrc->pMemDesc, pGpu);
+                status = heapReference(pGpu, pSrcMemDesc->pHeap, pMemorySrc->HeapOwner,
+                                       pSrcMemDesc);
+                NV_ASSERT_OK_OR_RETURN(status);
+            SLI_LOOP_END
+            break;
+        default:
+            NV_ASSERT_OR_RETURN(0, NV_ERR_INVALID_ARGUMENT);
+            break;
+    }
+
+    return status;
 }
 
 /*!
@@ -498,13 +541,14 @@ vidmemConstruct_IMPL
     MEMORY_ALLOCATION_REQUEST   *pAllocRequest         = &allocRequest;
     OBJGPU                      *pGpu                  = pMemory->pGpu;
     MemoryManager               *pMemoryManager        = GPU_GET_MEMORY_MANAGER(pGpu);
-    Heap                        *pHeap;
+    Heap                        *pHeap                 = NULL;
     NvBool                       bSubheap              = NV_FALSE;
     NvBool                       bRsvdHeap             = NV_FALSE;
     MEMORY_DESCRIPTOR           *pTopLevelMemDesc      = NULL;
     MEMORY_DESCRIPTOR           *pTempMemDesc          = NULL;
-    HWRESOURCE_INFO              hwResource;
+    HWRESOURCE_INFO              hwResource            = {0};
     RsClient                    *pRsClient             = pCallContext->pClient;
+    RmClient                    *pRmClient             = dynamicCast(pRsClient, RmClient);
     RsResourceRef               *pResourceRef          = pCallContext->pResourceRef;
     RsResourceRef               *pDeviceRef;
     Device                      *pDevice;
@@ -520,7 +564,8 @@ vidmemConstruct_IMPL
     FB_ALLOC_PAGE_FORMAT        *pFbAllocPageFormat    = NULL;
     NV_STATUS                    rmStatus              = NV_OK;
     KernelBus                   *pKernelBus            = GPU_GET_KERNEL_BUS(pGpu);
-    NvBool                       bUpdatePteKind        = NV_FALSE;
+
+    NV_ASSERT_OR_RETURN(pRmClient != NULL, NV_ERR_INVALID_CLIENT);
 
     NV_ASSERT_OK_OR_RETURN(
         refFindAncestorOfType(pResourceRef, classId(Device), &pDeviceRef));
@@ -542,22 +587,6 @@ vidmemConstruct_IMPL
         goto done;
     }
 
-    if (!FLD_TEST_DRF(OS32, _ATTR, _COMPR, _NONE, pAllocData->attr) &&
-        kbusIsStaticBar1Enabled(pGpu, pKernelBus))
-    {
-        if (!FLD_TEST_DRF(OS32, _ATTR, _PAGE_SIZE, _HUGE, pAllocData->attr))
-        {
-            // Override the attr to use 2MB page size
-            pAllocData->attr = FLD_SET_DRF(OS32, _ATTR, _PAGE_SIZE, _HUGE, pAllocData->attr);
-            pAllocData->attr2 = FLD_SET_DRF(OS32, _ATTR2, _PAGE_SIZE_HUGE, _DEFAULT, pAllocData->attr2);
-
-            NV_PRINTF(LEVEL_INFO,
-                      "Overrode the page size to 2MB on this compressed vidmem for the static bar1\n");
-        }
-
-        bUpdatePteKind = NV_TRUE;
-    }
-
     if (FLD_TEST_DRF(OS32, _ATTR, _PHYSICALITY, _DEFAULT, pAllocData->attr))
     {
         pAllocData->attr =
@@ -566,7 +595,7 @@ vidmemConstruct_IMPL
                             pAllocData->attr);
     }
 
-    NV_CHECK_OK_OR_RETURN(LEVEL_WARNING, stdmemValidateParams(pGpu, hClient, pAllocData));
+    NV_CHECK_OK_OR_RETURN(LEVEL_WARNING, stdmemValidateParams(pGpu, pRmClient, pAllocData));
     NV_CHECK_OR_RETURN(LEVEL_WARNING,
                        DRF_VAL(OS32, _ATTR, _LOCATION, pAllocData->attr) == NVOS32_ATTR_LOCATION_VIDMEM &&
                            !(pAllocData->flags & NVOS32_ALLOC_FLAGS_VIRTUAL),
@@ -631,6 +660,16 @@ vidmemConstruct_IMPL
         !memmgrIsScrubOnFreeEnabled(pMemoryManager) || bIsPmaAlloc || bSubheap || bRsvdHeap,
         NV_ERR_INVALID_STATE);
 
+    if (!FLD_TEST_DRF(OS32, _ATTR2, _ENABLE_LOCALIZED_MEMORY, _DEFAULT, pAllocData->attr2))
+    {
+        if (!bIsPmaAlloc)
+        {
+            // heap does not support localized allocations
+            rmStatus = NV_ERR_NOT_SUPPORTED;
+            goto done;
+        }
+    }
+
     // Get the allocation from PMA if enabled.
     if (bIsPmaAlloc)
     {
@@ -835,49 +874,51 @@ vidmemConstruct_IMPL
         SLI_LOOP_END;
     }
 
-    //
-    // Video memory is always locally transparently cached.  It does not require
-    // any cache managment.  Marked cached unconditionally.  Non-coherent peer
-    // caching is handled with an override at mapping time.
-    //
-    if (DRF_VAL(OS32, _ATTR2, _GPU_CACHEABLE, pAllocData->attr2) ==
-        NVOS32_ATTR2_GPU_CACHEABLE_DEFAULT)
     {
-        pAllocData->attr2 = FLD_SET_DRF(OS32, _ATTR2, _GPU_CACHEABLE, _YES,
-                                        pAllocData->attr2);
-    }
-    gpuCacheAttrib = NV_MEMORY_CACHED;
-
-    // ClientDB can set the pagesize for memdesc.
-    // With GPU SMMU mapping, this needs to be set on the SMMU memdesc.
-    // So SMMU allocation should happen before memConstructCommon()
-    // Eventaully SMMU allocation will be part of memdescAlloc().
-
-    //
-    // There are a few cases where the heap will return an existing
-    // memdesc.  Only update attributes if it is new.
-    //
-    // @todo attr tracking should move into heapAlloc
-    //
-    if (pTempMemDesc->RefCount == 1)
-    {
-        SLI_LOOP_START(SLI_LOOP_FLAGS_BC_ONLY | SLI_LOOP_FLAGS_IGNORE_REENTRANCY);
-        memdescSetGpuCacheAttrib(memdescGetMemDescFromGpu(pTopLevelMemDesc, pGpu), gpuCacheAttrib);
-        SLI_LOOP_END;
-
-
-        // An SMMU mapping will be added to FB allocations in the following cases:
-        // 1. RM clients forcing SMMU mapping via flags
-        //    GPU Arch verification with VPR is one such usecase.
-
-        if (FLD_TEST_DRF(OS32, _ATTR2, _SMMU_ON_GPU, _ENABLE, pAllocData->attr2))
+        //
+        // Video memory is always locally transparently cached.  It does not require
+        // any cache management.  Marked cached unconditionally.  Non-coherent peer
+        // caching is handled with an override at mapping time.
+        //
+        if (DRF_VAL(OS32, _ATTR2, _GPU_CACHEABLE, pAllocData->attr2) ==
+            NVOS32_ATTR2_GPU_CACHEABLE_DEFAULT)
         {
-            NV_ASSERT_FAILED("SMMU mapping allocation is not supported for ARMv7");
-            rmStatus = NV_ERR_NOT_SUPPORTED;
+            pAllocData->attr2 = FLD_SET_DRF(OS32, _ATTR2, _GPU_CACHEABLE, _YES,
+                                            pAllocData->attr2);
+        }
+        gpuCacheAttrib = NV_MEMORY_CACHED;
 
-            memdescFree(pTopLevelMemDesc);
-            memdescDestroy(pTopLevelMemDesc);
-            goto done;
+        // ClientDB can set the pagesize for memdesc.
+        // With GPU SMMU mapping, this needs to be set on the SMMU memdesc.
+        // So SMMU allocation should happen before memConstructCommon()
+        // Eventually SMMU allocation will be part of memdescAlloc().
+
+        //
+        // There are a few cases where the heap will return an existing
+        // memdesc.  Only update attributes if it is new.
+        //
+        // @todo attr tracking should move into heapAlloc
+        //
+        if (pTempMemDesc->RefCount == 1)
+        {
+            SLI_LOOP_START(SLI_LOOP_FLAGS_BC_ONLY | SLI_LOOP_FLAGS_IGNORE_REENTRANCY);
+            memdescSetGpuCacheAttrib(memdescGetMemDescFromGpu(pTopLevelMemDesc, pGpu), gpuCacheAttrib);
+            SLI_LOOP_END;
+
+
+            // An SMMU mapping will be added to FB allocations in the following cases:
+            // 1. RM clients forcing SMMU mapping via flags
+            //    GPU Arch verification with VPR is one such usecase.
+
+            if (FLD_TEST_DRF(OS32, _ATTR2, _SMMU_ON_GPU, _ENABLE, pAllocData->attr2))
+            {
+                NV_ASSERT_FAILED("SMMU mapping allocation is not supported for ARMv7");
+                rmStatus = NV_ERR_NOT_SUPPORTED;
+
+                memdescFree(pTopLevelMemDesc);
+                memdescDestroy(pTopLevelMemDesc);
+                goto done;
+            }
         }
     }
 
@@ -997,30 +1038,32 @@ vidmemConstruct_IMPL
         }
     }
 
-    if (bUpdatePteKind)
+    // No flags specified on the initial static BAR1 mapping 
+    rmStatus = kbusIncreaseStaticBar1Refcount_HAL(pGpu, pKernelBus,
+                   pMemory->pMemDesc, BUS_MAP_FB_FLAGS_NONE);
+
+    if (rmStatus == NV_OK)
     {
-        rmStatus = kbusUpdateStaticBar1VAMapping_HAL(pGpu, pKernelBus,
-                         pMemory->pMemDesc, 0,
-                         memdescGetSize(pMemory->pMemDesc), NV_FALSE);
-
-        if (rmStatus != NV_OK)
+        // nothing
+    }
+    else if (rmStatus == NV_ERR_NOT_SUPPORTED)
+    {
+        rmStatus = NV_OK;
+    }
+    else
+    {
+        if (pMemory->bRpcAlloc)
         {
-            if (pMemory->bRpcAlloc)
-            {
-                NV_STATUS status = NV_OK;
-                NV_RM_RPC_FREE(pGpu, hClient, hParent,
-                               pAllocRequest->hMemory, status);
-                NV_ASSERT(status == NV_OK);
-            }
-            memDestructCommon(pMemory);
-            memdescFree(pTopLevelMemDesc);
-            memdescDestroy(pTopLevelMemDesc);
-            pTopLevelMemDesc = NULL;
-            goto done;
+            NV_STATUS status = NV_OK;
+            NV_RM_RPC_FREE(pGpu, hClient, hParent,
+                           pAllocRequest->hMemory, status);
+            NV_ASSERT(status == NV_OK);
         }
-
-        memdescSetFlag(pMemory->pMemDesc,
-                       MEMDESC_FLAGS_RESTORE_PTE_KIND_ON_FREE, NV_TRUE);
+        memDestructCommon(pMemory);
+        memdescFree(pTopLevelMemDesc);
+        memdescDestroy(pTopLevelMemDesc);
+        pTopLevelMemDesc = NULL;
+        goto done;
     }
 
     pAllocData->size = sizeOut;
@@ -1063,15 +1106,26 @@ vidmemDestruct_IMPL
     Memory             *pMemory        = staticCast(pVideoMemory, Memory);
     OBJGPU             *pGpu           = pMemory->pGpu;
     MEMORY_DESCRIPTOR  *pMemDesc       = pMemory->pMemDesc;
+    MEMDESC_CUSTOM_HEAP customHeap;
 
     // Free any association of the memory with existing third-party p2p object
     CliUnregisterMemoryFromThirdPartyP2P(pMemory);
 
     memDestructCommon(pMemory);
 
+    //
+    // static BAR1: memory must be released from the static BAR1 mapping
+    // to restore the static BAR1 mapping to a default state since it can
+    // immediately be allocated by UVM after PMA free
+    //
+    (void)kbusDecreaseStaticBar1Refcount_HAL(pGpu,
+                    GPU_GET_KERNEL_BUS(pGpu), pMemDesc,
+                    NULL);
+
     // free the video memory based on how it was alloced ... a non-zero
     // heapOwner indicates it was heapAlloc-ed.
-    if (!memdescGetCustomHeap(pMemDesc))
+    customHeap = memdescGetCustomHeap(pMemDesc);
+    if (customHeap == MEMDESC_CUSTOM_HEAP_NONE)
     {
         MemoryManager      *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
         NvHandle            hClient        = RES_GET_CLIENT_HANDLE(pVideoMemory);
@@ -1271,6 +1325,21 @@ vidmemAllocResources
                          pAllocRequest->pPmaAllocInfo[subdeviceInst]->pageCount,
                          pAllocRequest->pPmaAllocInfo[subdeviceInst]->pageSize);
         }
+        if (!FLD_TEST_DRF(OS32, _ATTR2, _ENABLE_LOCALIZED_MEMORY, _DEFAULT, pVidHeapAlloc->attr2))
+        {
+            if (pMemoryManager->bLocalizedMemorySupported &&
+                !IS_MIG_ENABLED(pGpu))
+            {
+                memdescSetFlag(pAllocRequest->pMemDesc, MEMDESC_FLAGS_ALLOC_AS_LOCALIZED, NV_TRUE);
+                pAllocRequest->pMemDesc->localizedMask = pMemoryManager->localizedMask;
+            }
+            else
+            {
+                NV_PRINTF(LEVEL_ERROR, "Localized memory requested when localized memory not enabled\n");
+                status = NV_ERR_INVALID_ARGUMENT;
+                goto failed;
+            }
+        }
     }
     else
     {
